      subroutine cmpt_ort(fft_klh, hkl_max, fft_ann, t_analy,
     $     t_fftanaly, mem_cmpt, total_views, ctf_val, abs_ctf, disef,
     $     fno)

C     This is the key subroutine that performs the main function of
C     program POR - refinement of the particle origins and
C     orientations. The root process reads in the frames from the image
C     file and distributes them to the other processes. After all
C     processes have finished calculating the updated particle
C     information (orientation, origin, score), the root process
C     collects the results and writes out an updated parameter file with
C     the same base name as the original file and an updated suffix.
C
C     Arguments
C
C     fft_klh     - 3D FFT of model intent (in)
C     hkl_max     - array bound for interpolant array (in)
C     fft_ann     - array bound for fft_hkl (in)
C     t_analy     - cummulative timer (inout)
C     t_fftanaly  - cummulative timer for FFTs (inout)
C     mem_cmpt    - total memory allocated by routine (out)
C     total_views - cummulative count of number of images/views (inout)
C     ctf_val     - filtered, adjusted, weighted inverse of CTF (in)
C     abs_ctf     - filtered absolute value of CTF (in)
C     disef       - weighting function (in)
C     fno         - image file number (in)

      implicit none

      include 'mpif.h'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      
C     Subroutine arguments
      integer, intent(in) :: hkl_max, fft_ann, fno
      integer, intent(inout) :: total_views
      complex, intent(in) :: fft_klh(-fft_ann:fft_ann, 
     +     -fft_ann:fft_ann, 0:fft_ann)
      real, intent(in) :: ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in) :: abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in) :: disef( 0:nfc_2d_row, -nfc_2d_col:nfc_2d_col )
      double precision, intent(inout) :: t_analy, t_fftanaly
      double precision, intent(out) :: mem_cmpt
      
C     Local variables
      integer view, i, j, cc, max_node_eq
      integer :: ishift, maxshift = 10
      integer mem_size, start_view, end_view, num_views
      integer theta, phi, omega, ort_out, viewno, centeru, centerv
      integer min_angles(3), min_cen(2), sendcount(0:num_nodes),
     +     disp(0:num_nodes), minu, minv
      integer alloc_stat
      
      real r4_max
      real    angles_rad(3),angles_deg(3),center(2),imgmagfactor,
     +     num_annul,value,min_value, min_annul,
     +     du, dv
      real    rot_mat(3,3), 
     +     dist(-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle),
     +     distcen(-Ncenter:Ncenter,-Ncenter:Ncenter)
      real    orient(0:6, max_views), 
     +     orient_new(0:7, max_views/num_nodes+1),
     +     orient_out(0:7, max_views)
      
      real, allocatable :: rffta_omega(:), rffta(:), rffta_trns(:)
      real, allocatable :: pixels_array(:,:,:)
      complex, allocatable :: fc_2d(:,:)
      complex, allocatable :: fc_2d_intp(:,:)
      complex, allocatable :: fc_2d_bak(:,:)
      
      double precision wtime, t1, t2, time
      
      double precision :: mem_orient = 0.0D0 ,
     +     mem_rffta_omega = 0.0D0 , mem_rffta = 0.0D0 ,
     +     mem_rffta_trns = 0.0D0 , mem_fc_2d = 0.0D0 
      
      character(len=12) :: subname = 'cmpt_ort'

      r4_max = huge(r4_max)

C     Allocate memory for dynamic arrays
      mem_size        = ( 1 + nfc_2d_row ) * ( 1 + 2 * nfc_2d_col )
      mem_rffta_omega = dble(4 * nrffta_omega)
      mem_rffta       = dble( 4 * nrffta_row * nrffta_col )
      mem_rffta_trns  = dble( 4 * nrffta_row * nrffta_col )
      mem_orient      = dble(4*8*max_views)
      mem_fc_2d       = dble( 24 * mem_size )
      
      mem_cmpt = 3*mem_orient + mem_rffta_omega + 
     +     mem_rffta + mem_rffta_trns + mem_fc_2d
     +     + (2*Nangle+1)*(2*Nangle+1)*(2*Nangle+1)
     +     + (2*Ncenter+1)*(2*Ncenter+1)*(2*Ncenter+1)
      
      allocate(rffta_omega(nrffta_omega), rffta(nrffta_row*nrffta_col),
     +     rffta_trns(nrffta_row*nrffta_col), 
     +     fc_2d(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     fc_2d_intp(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     fc_2d_bak(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     pixels_array(npix_row, npix_col, view_perread),
     +     stat = alloc_stat)
      
      if(alloc_stat .gt. 0) then
         call error_stop(progname, subname, ' Memory allocation failed')
      endif

C     Initialize the array rffta_omega that is used by FFT library
      call vrffti(nrffta_col, rffta_omega)

      num_views = 0 
      viewno = 0 
      t1 = wtime()
      
C     Read particle parameters and distribute to all processes
      if( iam .eq. 0 ) then
         call readorient(orient, max_views, ort_file_name,
     +        num_views, npix_row, npix_col, oddeven)
         
         write(*,1001) iam, ort_file_name, num_views, max_views
 1001    format(/' Node = ',i2,':  Read orientation file: ',a/,
     +        12x,' Number of views     = ', i9, i9)
      endif
      
      call mpi_bcast(num_views, 1, mpi_integer,
     +     0, mpi_comm_world, mpi_err)
      if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +     'Error in calling MPI_BCAST num_views')
      
      call mpi_bcast(orient, 7*max_views, mpi_real,
     +     0, mpi_comm_world, mpi_err)
      if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +     'Error in calling MPI_BCAST orient')
      
      t2 = wtime()
      time = t2 - t1
      t_analy = t_analy + time
      
      total_views = total_views + num_views
      
      num_annul = real(nrffta_row)*angstrom_per_pix/res_max 
      min_annul = real(nrffta_row)*angstrom_per_pix/res_min 
      if (num_annul .ge. nrffta_row/2.) 
     +     call error_stop(progname,subname, 
     +     'Resolution is too high, smaller than 2*angstrom_per_pix')
      if (iam.eq.0) then
         print *,min_annul,num_annul,Dangle,Nangle,Dcenter,Ncenter
      endif
      
C     we seperate all the views by view_perread, 
C     each time we read at most view_perread * num_nodes 
C     pixels and deal with them
      
      DO j = 0, (num_views-1)/(view_perread*num_nodes)
         
         t1 = wtime()
         
         start_view = 1 + j * view_perread*num_nodes
         end_view   = start_view + view_perread*num_nodes - 1
         if (end_view .gt. num_views) end_view = num_views
         
C     the pix_file_name has been opened in aamain
C     Read the current end_view-start_view+1 pixel frames.
         
         call set_indices(num_nodes, start_view, end_view,
     +        frst_lst_view, max_node_eq )
C     CALL READ_FILES
C     ***************
C     processor 0 read pixels_array, then send to processor i
         do i = num_nodes - 1 , 0 , -1
            if (frst_lst_view(2, i) .GE. frst_lst_view(1, i)) then
               call read_files(pixels_array, frst_lst_view(1,i),
     +              frst_lst_view(2,i), i, orient)
            endif
         enddo
         
         t2 = wtime()
         time = t2 - t1
         t_analy = t_analy + time
         
C     do with each view and pixel in  per read
         
         start_view = frst_lst_view(1,iam)
         end_view   = frst_lst_view(2,iam)
         
         do view = start_view, end_view
C     
C     Get the orientation angles and center of next pixel frame
C     Also changes OMEGA from the range (0,360) deg to (-90,270) deg.
C     Because the orient used by BIO not consistent with defination 
C     CALL GETORIENT
C     **************
            call getorient( orient, max_views, view,
     +           angles_rad, angles_deg, center, imgmagfactor)
C     adjust the imgmagfactor according to its own pixel_size
            imgmagfactor = imgmagfactor * PIXSIZ_OUT / angstrom_per_pix
C     
C------------------------------------------------------------
C     Compute Discrete Fourier Transform of pixel values:
C     CALL FFTANALY
C     *************
            t1 = wtime()
            
            call fft_2dfft( pixels_array(1, 1, view-start_view+1),
     +           npix_row, npix_col, center,
     +           nrffta_omega, rffta_omega, nrffta_row, 
     +           nrffta_col, nfc_2d_row, 
     +           nfc_2d_col, fc_2d )
            du = center(1) - real(int(center(1)))
            dv = center(2) - real(int(center(2)))
            call shift_fft(fc_2d, nfc_2d_row, nfc_2d_col, du, dv)
            
C     backup fc_2d
            fc_2d_bak = fc_2d  ! f90 array syntax
            fc_2d_intp = 0.    ! f90 array syntax

            t2 = wtime()
            time = t2 - t1
            t_fftanaly = t_fftanaly + time

            min_value = R4_MAX


C     ----------- Refine particle orientation -----------
C
C     Search a region of orientation space, centered on the current
C     orientation, for a new model projection that has the best
C     agreement (lowest score) with the particle image. The number of
C     steps taken in each direction for each angle is nangle. If the new
C     optimal orientation is more than nangle/2 steps from the starting
C     point in any direction, perform another search centered on this
C     new orientation. To minimize the number of comparisons that need
C     to be calculated, the array DIST containing the scores is
C     recentered so that previously calculated scores can be reused. The
C     outer loop over ishift sets a limit on the number of times that
C     search can be restarted.
C
C     Note that in this convention, the angle omega has to be set to
C     (360-omega-90), or equivalently, -(omega+90).


C     Apply CTF correction (f90 array syntax)
            IF(CTFMODE.GT.0 .and. CTFMODE.LE.4)  then
               fc_2d = fc_2d * ctf_val
            endif

C     Initialize dist array to very large value (f90 array syntax)
            dist = R4_MAX

            do ishift = 1, maxshift
               
               min_angles(1) = 0
               min_angles(2) = 0
               min_angles(3) = 0
               
C     Search orientation space
               do omega = -Nangle, Nangle  
                  do phi   = -Nangle, Nangle  
                     do theta = -Nangle, Nangle  
                        
                        angles_rad(1)= (Dangle*theta+angles_deg(1))*
     +                       deg_to_rad
                        angles_rad(2)= (Dangle*phi+angles_deg(2))*
     +                       deg_to_rad
                        angles_rad(3)=-(Dangle*omega+angles_deg(3)+90.)*
     +                       deg_to_rad

                        if (dist(theta,phi,omega).GE.R4_MAX-1) then
                           call setrotmat( angles_rad, rot_mat)
                           call interpl_3d(fft_klh, hkl_max, 
     +                          min_annul, num_annul,
     +                          rot_mat, fc_2d_intp, imgmagfactor, 
     +                          fft_ann) 
                           call imgcomport(fc_2d_intp, hkl_max, 
     +                          fc_2d, disef, abs_ctf, min_annul, 
     +                          num_annul, value, funcmode)
                           dist(theta, phi, omega) = value
                        
                           if ( value .lt. min_value) then
                              min_value = value
                              min_angles(1) = theta
                              min_angles(2) = phi
                              min_angles(3) = omega
                           endif
                        endif
                     enddo
                  enddo
               enddo
               
               print *, iam, 'min_value=', min_value,'<=',
     +              min_angles(1), min_angles(2), min_angles(3)

               angles_deg(1) = angles_deg(1) + min_angles(1) * Dangle
               angles_deg(2) = angles_deg(2) + min_angles(2) * Dangle
               angles_deg(3) = angles_deg(3) + min_angles(3) * Dangle
               
C     If the new optimal orientation is sufficiently far from the
C     starting orientation, recenter the array DIST and start a new
C     search. Otherwise, finished with orientation refinement.

               if ( abs(real(min_angles(1))) .gt. Nangle*0.5 .or. 
     +              abs(real(min_angles(2))) .gt. Nangle*0.5 .or.
     +              abs(real(min_angles(3))) .gt. Nangle*0.5 ) then
                  call recenter_array_3d(dist,Nangle,min_angles,R4_MAX)
               else
                  exit
               endif
            enddo

C     ----------- Check handedness of orientation -----------
C
C     Compare current orientation with the corresponding orientation
C     with the opposite handedness (theta, -phi, omega+180). Switch
C     handedness of the opposite handed orientation has a better match
C     with the model.

            angles_rad(1)= angles_deg(1)*deg_to_rad
            angles_rad(2)= - angles_deg(2)*deg_to_rad
            angles_rad(3)= -(angles_deg(3)-90.)*deg_to_rad

            call setrotmat( angles_rad, rot_mat)
            call interpl_3d(fft_klh, hkl_max,
     +           min_annul, num_annul,
     +           rot_mat, fc_2d_intp, imgmagfactor, fft_ann) 
            call imgcomport(fc_2d_intp, hkl_max, 
     +           fc_2d, disef, abs_ctf, min_annul, num_annul,
     +           value, funcmode)
            if ( value .lt. min_value) then
               print *, iam, 'min_value=', value
               min_value = value
               angles_deg(2) = - angles_deg(2)
               angles_deg(3) = angles_deg(3) + 180.0
            endif

C     ----------- Return orientation to asymmetric unit -----------
C
            if ( symm_code .gt. 1) then
               call to_asym_unit(symm_code,
     +              angles_deg(1), angles_deg(2), angles_deg(3))
            endif


C     ----------- Refine particle origin -----------
C
C     Search a region of origin space, centered on the current origin,
C     for a new model projection that has the best agreement (lowest
C     score) with the particle image. The number of steps taken in each
C     direction for each coordinate is ncenter. If the new optimal
C     origin is more than ncenter/2 steps from the starting point in any
C     direction, perform another search centered on this new origin. To
C     minimize the number of comparisons that need to be calculated, the
C     array DISTCEN containing the scores is recentered so that
C     previously calculated scores can be reused. The outer loop over
C     ishift sets a limit on the number of times that search can be
C     restarted.

            angles_rad(1) = angles_deg(1)*deg_to_rad
            angles_rad(2) = angles_deg(2)*deg_to_rad
            angles_rad(3) = -(angles_deg(3)+90.)*deg_to_rad
            
            call setrotmat( angles_rad, rot_mat)
            call interpl_3d(fft_klh, hkl_max,
     +           min_annul, num_annul,
     +           rot_mat, fc_2d_intp, imgmagfactor, fft_ann) 
            
            distcen = R4_MAX    ! f90 array syntax
            
            min_value = R4_MAX
            minu      = 0
            minv      = 0

            do ishift = 1, maxshift
               
               min_cen(1) = 0
               min_cen(2) = 0
               
               do centerv = -Ncenter, Ncenter 
                  dv = Dcenter * (centerv + minv)
                  do centeru = -Ncenter, Ncenter 
                     if (distcen(centeru,centerv).GE.R4_MAX-1) then
                        du = Dcenter * (centeru + minu)
                        fc_2d = fc_2d_bak
                        call shift_fft(fc_2d,
     +                       nfc_2d_row, nfc_2d_col, du, dv)

C     Apply CTF correction
                        IF(CTFMODE.GT.0 .and. CTFMODE.LE.4)  then
                           fc_2d = fc_2d * ctf_val !f90 array syntax
                        endif
                        
                        call imgcompcet(fc_2d_intp, hkl_max, 
     +                       fc_2d, disef, abs_ctf, min_annul,
     +                       num_annul, value)
                        distcen(centeru,centerv) = value
                        if ( value .lt. min_value) then
                           min_value = value
                           min_cen(1) = centeru
                           min_cen(2) = centerv
                        endif
                     endif
                  enddo
               enddo
            
               print *, iam, 'min_value=', min_value,'center=(',
     +              min_cen(1), min_cen(2), ')'
               
C     If the new optimal origin is sufficiently far from the starting
C     origin, recenter the array DISTCEN and start a new
C     search. Otherwise, finished with origin refinement.

               if ( abs(real(min_cen(1))) .gt. Ncenter*0.5 .or. 
     +              abs(real(min_cen(2))) .gt. Ncenter*0.5 ) then
                  minu = minu + min_cen(1) 
                  minv = minv + min_cen(2)
                  call recenter_array_2d(distcen,Ncenter,min_cen,R4_MAX)
               else
                  exit
               endif

            enddo

            du = (minu+min_cen(1))*Dcenter
            dv = (minv+min_cen(2))*Dcenter

            fc_2d = fc_2d_bak
            call shift_fft(fc_2d, nfc_2d_row, nfc_2d_col, du, dv)

C     ctf correction

            IF(CTFMODE.GT.0 .and. CTFMODE.LE.4)  then
               fc_2d = fc_2d * ctf_val !f90 array syntax
            endif
            
C     compute the CC value for the DFT and best CUT
            call imgcomport(fc_2d_intp, hkl_max, 
     +           fc_2d, disef, abs_ctf, min_annul, num_annul,
     +           min_value, funcmode)
            
C     update new center and CC value
            viewno               = viewno +1
            orient_new(0,viewno) = orient(0, view)
            orient_new(1,viewno) = angles_deg(1)
            orient_new(2,viewno) = angles_deg(2)
            orient_new(3,viewno) = angles_deg(3)
            orient_new(4,viewno) = orient(4,view)+ du
            orient_new(5,viewno) = orient(5,view)+ dv
            orient_new(6,viewno) = orient(6,view)
            orient_new(7,viewno) = min_value
            
            write(*, '(i6, 5(f10.3), f6.3)')
     +           nint(orient(0, view)), 
     +           (orient(cc, view),cc=1,6)
            write(*, '(i6, 5(f10.3), f6.3, f10.3)')
     +           nint(orient_new(0, viewno)), 
     +           (orient_new(cc, viewno),cc=1,7)
            
         enddo
      enddo             
      
C     Gather results onto root process
      call set_indices(num_nodes, 1, num_views, frst_lst_view, 
     +     max_node_eq )

      do i = 0, num_nodes-1
         sendcount(i)=(frst_lst_view(2,i)-frst_lst_view(1,i)+1)*8
         if (i.eq.0) then 
            disp(0) = 0
         else
            disp(i) = disp(i-1) + sendcount(i-1)
         endif
      enddo
      
      call mpi_gatherv(orient_new,sendcount(iam),mpi_real,orient_out,
     +     sendcount,disp,mpi_real,0,mpi_comm_world,mpi_err)
      if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +     'Error in calling MPI_gatherv orient_out')
      

C     Write out new particle parameter file from root process
      if (iam.eq.0) then
         call increment_fname(ort_file_name, ort_out_name)
         ort_out = 11
         OPEN (UNIT=ort_out, FILE=ort_out_name)
         write(ort_out,'(A)') pix_file_name
         write(ort_out,1010) PIXSIZS(fno), units, VOLTS(fno), 
     +        AMPFACS(fno), DELF_MAJS(fno), DELF_MINS(fno), 
     +        ANG_MAJS(fno), CS_COEFFS(fno)
 1010    format(1X,F8.4,',',I2,',',F10.1,',',4(F10.4,','),F8.2) 

         do view = 1, num_views
            write(ort_out,1020) nint(orient_out(0, view)),
     +           (orient_out(cc, view),cc=1,7)
         enddo
         close(UNIT=ort_out)
 1020    format(I7,',',4(F8.3,','),F7.3,',',F6.3,',',
     +        F10.3,',',' ,',' ,')
      endif
      
C     Free memory allocated in this subroutine.
      deallocate(rffta_omega, rffta, rffta_trns, fc_2d,
     +     fc_2d_intp, fc_2d_bak, pixels_array)
      
      return
      end subroutine cmpt_ort
